from __future__ import annotations

import json
from typing import Literal, Optional
from pydantic import BaseModel, Field

from openreward.environments import Toolset, tool, ToolOutput, TextBlock


# ===== Pydantic Parameter Models =====

class CreateSpreadsheetParams(BaseModel):
    file_path: str = Field(..., description="Path where the new spreadsheet will be created")
    sheet_name: str = Field("Sheet1", description="Name of the initial worksheet")


class DeleteSpreadsheetParams(BaseModel):
    file_path: str = Field(..., description="Path to the Excel spreadsheet to delete")


class ListTabsParams(BaseModel):
    file_path: str = Field(..., description="Path to the Excel spreadsheet")


class AddTabParams(BaseModel):
    file_path: str = Field(..., description="Path to the Excel spreadsheet")
    sheet_name: str = Field(..., description="Name for the new worksheet")
    position: int | None = Field(None, description="Position index (0-based), None=append at end")


class DeleteTabParams(BaseModel):
    file_path: str = Field(..., description="Path to the Excel spreadsheet")
    sheet_name: str = Field(..., description="Name of the worksheet to delete")


class ReadTabParams(BaseModel):
    file_path: str = Field(..., description="Path to the Excel spreadsheet")
    sheet_name: str = Field(..., description="Name of the worksheet to read")
    max_rows: int | None = Field(None, description="Maximum number of rows to read")
    max_cols: int | None = Field(None, description="Maximum number of columns to read")


class ReadCsvParams(BaseModel):
    csv_path: str = Field(..., description="Path to the CSV file")
    max_rows: int | None = Field(None, description="Maximum number of rows to read")


class EditSpreadsheetParams(BaseModel):
    file_path: str = Field(..., description="Path to the Excel spreadsheet")
    sheet_name: str = Field(..., description="Name of the worksheet")
    cell_reference: str = Field(..., description="Cell reference (e.g., 'A1', 'B5')")
    value: str | int | float = Field(..., description="Value to write to the cell")


class AddContentTextParams(BaseModel):
    file_path: str = Field(..., description="Path to the Excel spreadsheet")
    sheet_name: str = Field(..., description="Name of the worksheet")
    start_cell: str = Field(..., description="Starting cell reference (e.g., 'A1')")
    data: list[list] = Field(..., description="2D array of values to write [row][col]")


class DeleteContentCellParams(BaseModel):
    file_path: str = Field(..., description="Path to the Excel spreadsheet")
    sheet_name: str = Field(..., description="Name of the worksheet")
    cell_reference: str = Field(..., description="Cell reference to clear (e.g., 'A1')")


class CreateChartParams(BaseModel):
    file_path: str = Field(..., description="Path to the Excel spreadsheet")
    sheet_name: str = Field(..., description="Name of the worksheet")
    chart_type: Literal["bar", "column", "line", "pie", "scatter", "area"] = Field(..., description="Type of chart")
    data_range: str = Field(..., description="Data range for chart (e.g., 'A1:B10')")
    chart_position: str = Field(..., description="Cell reference for chart top-left corner (e.g., 'D1')")
    title: str | None = Field(None, description="Chart title")
    x_axis_title: str | None = Field(None, description="X-axis label")
    y_axis_title: str | None = Field(None, description="Y-axis label")


# ===== Excel Toolset Class =====

class ExcelToolset(Toolset):
    """Toolset providing Excel spreadsheet manipulation tools via openpyxl executed in sandbox"""

    @tool
    async def excel_create_spreadsheet(self, params: CreateSpreadsheetParams) -> ToolOutput:
        """Create a new Excel workbook with initial worksheet"""
        sheet_name_escaped = params.sheet_name.replace('\\', '\\\\').replace('"', '\\"')

        script = f'''
import json
from openpyxl import Workbook

try:
    wb = Workbook()
    ws = wb.active
    ws.title = "{sheet_name_escaped}"

    wb.save("{params.file_path}")

    result = {{
        "success": True,
        "file_path": "{params.file_path}",
        "sheet_name": "{sheet_name_escaped}",
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Spreadsheet created successfully at {params.file_path}")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_delete_spreadsheet(self, params: DeleteSpreadsheetParams) -> ToolOutput:
        """Delete an Excel spreadsheet file from sandbox"""
        output, exit_code = await self.sandbox.run(f"rm {params.file_path}", max_bytes=1_000_000)

        if exit_code == 0:
            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Spreadsheet deleted successfully: {params.file_path}")],
                metadata={"file_path": params.file_path, "deleted": True},
                reward=0.0,
                finished=False,
            )
        else:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Failed to delete: {output}")],
                metadata={"file_path": params.file_path, "deleted": False, "error": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_list_tabs_in_spreadsheet(self, params: ListTabsParams) -> ToolOutput:
        """List all worksheet/tab names in the workbook"""
        script = f'''
import json
from openpyxl import load_workbook

try:
    wb = load_workbook("{params.file_path}")
    sheet_names = wb.sheetnames

    result = {{
        "success": True,
        "file_path": "{params.file_path}",
        "sheet_count": len(sheet_names),
        "sheet_names": sheet_names,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            # Build summary
            summary = f"Spreadsheet: {params.file_path}\n"
            summary += f"Total sheets: {result['sheet_count']}\n\n"
            summary += "Sheets:\n"
            for idx, name in enumerate(result['sheet_names']):
                summary += f"  [{idx}] {name}\n"

            return ToolOutput(
                blocks=[TextBlock(text=summary)],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_add_tab(self, params: AddTabParams) -> ToolOutput:
        """Add a new worksheet to existing workbook"""
        sheet_name_escaped = params.sheet_name.replace('\\', '\\\\').replace('"', '\\"')
        position_code = f"{params.position}" if params.position is not None else "None"

        script = f'''
import json
from openpyxl import load_workbook

try:
    wb = load_workbook("{params.file_path}")

    # Create new sheet at specified position or end
    position = {position_code}
    if position is not None:
        wb.create_sheet(title="{sheet_name_escaped}", index=position)
    else:
        wb.create_sheet(title="{sheet_name_escaped}")

    wb.save("{params.file_path}")

    result = {{
        "success": True,
        "file_path": "{params.file_path}",
        "sheet_name": "{sheet_name_escaped}",
        "sheet_count": len(wb.sheetnames),
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Sheet '{params.sheet_name}' added successfully (total: {result['sheet_count']} sheets)")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_delete_tab(self, params: DeleteTabParams) -> ToolOutput:
        """Remove a worksheet from workbook"""
        sheet_name_escaped = params.sheet_name.replace('\\', '\\\\').replace('"', '\\"')

        script = f'''
import json
from openpyxl import load_workbook

try:
    wb = load_workbook("{params.file_path}")

    if "{sheet_name_escaped}" not in wb.sheetnames:
        print(json.dumps({{
            "success": False,
            "error": f"Sheet '{{sheet_name_escaped}}' not found. Available sheets: {{wb.sheetnames}}"
        }}))
    else:
        ws = wb["{sheet_name_escaped}"]
        wb.remove(ws)
        wb.save("{params.file_path}")

        result = {{
            "success": True,
            "file_path": "{params.file_path}",
            "deleted_sheet": "{sheet_name_escaped}",
            "remaining_sheets": len(wb.sheetnames),
        }}
        print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Sheet '{params.sheet_name}' deleted successfully ({result['remaining_sheets']} sheets remaining)")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_read_tab(self, params: ReadTabParams) -> ToolOutput:
        """Read data from a specific worksheet"""
        sheet_name_escaped = params.sheet_name.replace('\\', '\\\\').replace('"', '\\"')
        max_rows = params.max_rows if params.max_rows else 999999
        max_cols = params.max_cols if params.max_cols else 999999

        script = f'''
import json
import datetime
from openpyxl import load_workbook

def default_serializer(obj):
    if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
        return obj.isoformat()
    if isinstance(obj, datetime.timedelta):
        return str(obj)
    raise TypeError(f"Object of type {{type(obj).__name__}} is not JSON serializable")

try:
    wb = load_workbook("{params.file_path}")

    if "{sheet_name_escaped}" not in wb.sheetnames:
        print(json.dumps({{
            "success": False,
            "error": f"Sheet '{{sheet_name_escaped}}' not found. Available sheets: {{wb.sheetnames}}"
        }}))
    else:
        ws = wb["{sheet_name_escaped}"]

        # Read data from worksheet
        data = []
        for row_idx, row in enumerate(ws.iter_rows(values_only=True)):
            if row_idx >= {max_rows}:
                break
            row_data = list(row[:min(len(row), {max_cols})])
            data.append(row_data)

        result = {{
            "success": True,
            "file_path": "{params.file_path}",
            "sheet_name": "{sheet_name_escaped}",
            "rows": len(data),
            "cols": len(data[0]) if data else 0,
            "data": data,
        }}
        print(json.dumps(result, default=default_serializer))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            # Build summary with preview
            summary = f"Read sheet '{params.sheet_name}' from {params.file_path}\n"
            summary += f"Dimensions: {result['rows']} rows × {result['cols']} columns\n\n"
            summary += "Data:\n"
            for row_idx, row in enumerate(result['data']):
                row_str = ", ".join([str(cell) if cell is not None else "(empty)" for cell in row])
                summary += f"  Row {row_idx + 1}: {row_str}\n"

            return ToolOutput(
                blocks=[TextBlock(text=summary)],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_read_csv(self, params: ReadCsvParams) -> ToolOutput:
        """Read CSV file and convert to Excel format data structure"""
        max_rows = params.max_rows if params.max_rows else 999999

        script = f'''
import json
import csv

try:
    data = []
    with open("{params.csv_path}", 'r') as f:
        reader = csv.reader(f)
        for row_idx, row in enumerate(reader):
            if row_idx >= {max_rows}:
                break
            data.append(row)

    result = {{
        "success": True,
        "csv_path": "{params.csv_path}",
        "rows": len(data),
        "cols": len(data[0]) if data else 0,
        "data": data,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            # Build summary with preview
            summary = f"Read CSV file: {params.csv_path}\n"
            summary += f"Dimensions: {result['rows']} rows × {result['cols']} columns\n\n"
            summary += "Data:\n"
            for row_idx, row in enumerate(result['data']):
                row_str = ", ".join(row)
                summary += f"  Row {row_idx + 1}: {row_str}\n"

            return ToolOutput(
                blocks=[TextBlock(text=summary)],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_edit_spreadsheet(self, params: EditSpreadsheetParams) -> ToolOutput:
        """Modify existing cell value in a worksheet"""
        sheet_name_escaped = params.sheet_name.replace('\\', '\\\\').replace('"', '\\"')

        # Handle different value types
        if isinstance(params.value, str):
            value_str = params.value.replace('\\', '\\\\').replace('"', '\\"')
            value_code = f'"{value_str}"'
        else:
            value_code = str(params.value)

        script = f'''
import json
from openpyxl import load_workbook

try:
    wb = load_workbook("{params.file_path}")

    if "{sheet_name_escaped}" not in wb.sheetnames:
        print(json.dumps({{
            "success": False,
            "error": f"Sheet '{{sheet_name_escaped}}' not found. Available sheets: {{wb.sheetnames}}"
        }}))
    else:
        ws = wb["{sheet_name_escaped}"]
        old_value = ws["{params.cell_reference}"].value

        ws["{params.cell_reference}"] = {value_code}

        wb.save("{params.file_path}")

        result = {{
            "success": True,
            "file_path": "{params.file_path}",
            "sheet_name": "{sheet_name_escaped}",
            "cell_reference": "{params.cell_reference}",
            "old_value": str(old_value) if old_value is not None else None,
            "new_value": str(ws["{params.cell_reference}"].value),
        }}
        print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Cell {params.cell_reference} updated in sheet '{params.sheet_name}'\nNew value: {result['new_value']}")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_add_content_text(self, params: AddContentTextParams) -> ToolOutput:
        """Write data to a range of cells (batch operation)"""
        sheet_name_escaped = params.sheet_name.replace('\\', '\\\\').replace('"', '\\"')
        data_json = json.dumps(params.data)

        script = f'''
import json
from openpyxl import load_workbook
from openpyxl.utils import get_column_letter, column_index_from_string

try:
    wb = load_workbook("{params.file_path}")

    if "{sheet_name_escaped}" not in wb.sheetnames:
        print(json.dumps({{
            "success": False,
            "error": f"Sheet '{{sheet_name_escaped}}' not found. Available sheets: {{wb.sheetnames}}"
        }}))
    else:
        ws = wb["{sheet_name_escaped}"]
        data = {data_json}

        # Parse start cell (e.g., "A1" -> row=1, col=1)
        cell_ref = "{params.start_cell}"
        col_letter = ""
        row_num = ""
        for char in cell_ref:
            if char.isalpha():
                col_letter += char
            else:
                row_num += char

        start_row = int(row_num)
        start_col = column_index_from_string(col_letter)

        # Write data
        for row_idx, row_data in enumerate(data):
            for col_idx, value in enumerate(row_data):
                ws.cell(row=start_row + row_idx, column=start_col + col_idx, value=value)

        wb.save("{params.file_path}")

        result = {{
            "success": True,
            "file_path": "{params.file_path}",
            "sheet_name": "{sheet_name_escaped}",
            "start_cell": "{params.start_cell}",
            "rows_written": len(data),
            "cols_written": len(data[0]) if data else 0,
        }}
        print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Data written to sheet '{params.sheet_name}' starting at {params.start_cell}\nDimensions: {result['rows_written']} rows × {result['cols_written']} columns")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_delete_content_cell(self, params: DeleteContentCellParams) -> ToolOutput:
        """Clear cell content (sets to None)"""
        sheet_name_escaped = params.sheet_name.replace('\\', '\\\\').replace('"', '\\"')

        script = f'''
import json
from openpyxl import load_workbook

try:
    wb = load_workbook("{params.file_path}")

    if "{sheet_name_escaped}" not in wb.sheetnames:
        print(json.dumps({{
            "success": False,
            "error": f"Sheet '{{sheet_name_escaped}}' not found. Available sheets: {{wb.sheetnames}}"
        }}))
    else:
        ws = wb["{sheet_name_escaped}"]
        old_value = ws["{params.cell_reference}"].value

        ws["{params.cell_reference}"] = None

        wb.save("{params.file_path}")

        result = {{
            "success": True,
            "file_path": "{params.file_path}",
            "sheet_name": "{sheet_name_escaped}",
            "cell_reference": "{params.cell_reference}",
            "deleted_value": str(old_value) if old_value is not None else "(empty)",
        }}
        print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Cell {params.cell_reference} cleared in sheet '{params.sheet_name}'\nDeleted value: {result['deleted_value']}")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def excel_create_chart(self, params: CreateChartParams) -> ToolOutput:
        """Create a chart in the worksheet"""
        sheet_name_escaped = params.sheet_name.replace('\\', '\\\\').replace('"', '\\"')
        title_escaped = params.title.replace('\\', '\\\\').replace('"', '\\"') if params.title else ""
        x_axis_escaped = params.x_axis_title.replace('\\', '\\\\').replace('"', '\\"') if params.x_axis_title else ""
        y_axis_escaped = params.y_axis_title.replace('\\', '\\\\').replace('"', '\\"') if params.y_axis_title else ""

        # Map chart type to openpyxl chart class
        chart_type_map = {
            "bar": "BarChart",
            "column": "BarChart",  # BarChart with different direction
            "line": "LineChart",
            "pie": "PieChart",
            "scatter": "ScatterChart",
            "area": "AreaChart",
        }
        chart_class = chart_type_map.get(params.chart_type, "BarChart")

        # Construct the full reference string with sheet name
        # Quote sheet name if it contains spaces or special characters
        if ' ' in params.sheet_name or any(c in params.sheet_name for c in ['!', "'", '"']):
            full_data_ref = f"'{params.sheet_name}'!{params.data_range}"
        else:
            full_data_ref = f"{params.sheet_name}!{params.data_range}"

        script = f'''
import json
from openpyxl import load_workbook
from openpyxl.chart import BarChart, LineChart, PieChart, ScatterChart, AreaChart, Reference

try:
    wb = load_workbook("{params.file_path}")

    if "{sheet_name_escaped}" not in wb.sheetnames:
        print(json.dumps({{
            "success": False,
            "error": f"Sheet '{{sheet_name_escaped}}' not found. Available sheets: {{wb.sheetnames}}"
        }}))
    else:
        ws = wb["{sheet_name_escaped}"]

        # Create chart
        chart = {chart_class}()

        # Use full reference string with sheet name
        data = Reference(ws, range_string="{full_data_ref}")

        chart.add_data(data, titles_from_data=True)

        # Set chart properties
        if "{title_escaped}":
            chart.title = "{title_escaped}"
        if "{x_axis_escaped}":
            chart.x_axis.title = "{x_axis_escaped}"
        if "{y_axis_escaped}":
            chart.y_axis.title = "{y_axis_escaped}"

        # For column chart, set type to column
        if "{params.chart_type}" == "column":
            chart.type = "col"

        # Add chart to worksheet
        ws.add_chart(chart, "{params.chart_position}")

        wb.save("{params.file_path}")

        result = {{
            "success": True,
            "file_path": "{params.file_path}",
            "sheet_name": "{sheet_name_escaped}",
            "chart_type": "{params.chart_type}",
            "data_range": "{params.data_range}",
            "chart_position": "{params.chart_position}",
            "title": "{title_escaped}" if "{title_escaped}" else None,
        }}
        print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            chart_desc = f"{params.chart_type.capitalize()} chart"
            if params.title:
                chart_desc += f" '{params.title}'"

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ {chart_desc} created in sheet '{params.sheet_name}' at {params.chart_position}")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Mapping, Optional, Sequence, Literal, Union

from typing_extensions import TypeAliasType

JSONValue = TypeAliasType("JSONValue", Union[Mapping[str, Any], Sequence[Any], str, int, float, bool, None])
JSONObject = Mapping[str, JSONValue]

@dataclass
class Server:
    name: str

@dataclass
class Environment:
    server_name: str
    environment_name: str
    namespace: str = "matrix"

    @property
    def deployment_name(self) -> str:
        """Get the full deployment identifier in namespace/server_name format."""
        if "/" in self.server_name:
            # Already contains namespace
            return self.server_name
        return f"{self.namespace}/{self.server_name}"

@dataclass
class Task:
    server_name: str
    environment_name: str
    task_spec: JSONObject
    namespace: Optional[str]

    @property
    def deployment_name(self) -> str:
        """Get the full deployment identifier in namespace/server_name format."""
        if self.namespace is None:
            return self.server_name
        else:
            return f"{self.namespace}/{self.server_name}"

@dataclass
class ToolSpec:
    name: str
    description: str
    input_schema: Optional[JSONObject]

@dataclass
class TextBlock:
    text: str
    detail: Optional[JSONObject] = None
    type: Literal["text"] = "text"

@dataclass
class ImageBlock:
    data: str
    mimeType: str
    detail: Optional[JSONObject] = None
    type: Literal["image"] = "image"


@dataclass
class ToolOutput:
    blocks: list[Union[TextBlock, ImageBlock]]
    metadata: Optional[JSONObject] = None
    reward: Optional[float] = None
    finished: bool = False

class ToolCallError(Exception):
    pass

from openreward.api._session.http import AuthenticationError as AuthenticationError  # noqa: F401

Provider = Literal[
    "openai",
    "anthropic",
    "google",
    "openrouter"
]


"""Base class for toolsets that provide reusable collections of tools."""

from abc import ABC
from typing import Any


class Toolset(ABC):
    """
    Optional base class for toolsets. Provides common utilities for
    sandbox-based toolsets.

    Toolsets are collections of related tools that can be easily reused
    across different environments. They follow the same @tool decorator
    pattern as Environment methods.

    Usage:
        from openreward.environments import Toolset, tool, ToolOutput

        class MyToolset(Toolset):
            @tool
            async def my_tool(self) -> ToolOutput:
                # Use self.sandbox to interact with sandbox
                output, code = await self.sandbox.run("ls")
                return ToolOutput(blocks=[TextBlock(text=output)])

        # In your environment:
        class MyEnv(Environment):
            toolsets = [MyToolset]

            def __init__(self, task_spec, secrets):
                super().__init__(task_spec, secrets)
                self.sandbox = ...  # MyToolset will access this automatically
    """

    def __init__(self, env: Any, sandbox_attr: str = "sandbox"):
        """
        Initialize toolset with environment dependencies.

        Args:
            env: The environment instance that owns this toolset
            sandbox_attr: Name of sandbox attribute on env (default: "sandbox").
                         The toolset will look for this attribute and store it
                         as self.sandbox for easy access.

        Raises:
            AttributeError: If the specified sandbox_attr doesn't exist on env
        """
        self.env = env
        self.sandbox = getattr(env, sandbox_attr)



import asyncio
import base64
import inspect
import json
import time
import traceback
import uuid
from contextlib import asynccontextmanager, suppress
from typing import Literal, Optional, Union

import structlog
import uvicorn
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse, RedirectResponse
from fastapi.routing import APIRoute
from sse_starlette import EventSourceResponse
from starlette.types import ASGIApp, Receive, Scope, Send

from openreward._version import __version__
from openreward.log_utils import get_logger as _get_logger, setup_logging, OPENREWARD_USE_STRUCTURED_LOGS
from .environment import Environment
from .reconnect import sse_task_stream
from .types import Blocks, CreateSession, GetTask, GetTaskRange, ListTasks, ListToolsOutput, NumTasks, Split, ToolCall
from .utils import maybe_await

logger = _get_logger("openreward.environments.server")


def _get_env_map(environments: list[type[Environment]]) -> dict[str, type[Environment]]:
    env_map = {}
    for env in environments:
        # Always treat env as a class
        cls = env if inspect.isclass(env) else type(env)
        # Ensure we don't pass non-type or overloads to issubclass
        if not (isinstance(cls, type) and issubclass(cls, Environment)):
            raise TypeError(f"{cls!r} is not Environment")
        key: str = cls.name().lower()
        if key in env_map:
            raise ValueError(f"duplicate env {key}")
        env_map[key] = cls
    return env_map

def _get_env_cls(env_map: dict[str, type[Environment]], env_name: str) -> type[Environment]:
    env_cls = env_map.get(env_name.lower())
    if env_cls is None:
        raise HTTPException(404, f"unknown environment {env_name!r}")
    return env_cls


def _convert_to_split(split: Union[str, Split]) -> Split:
    if isinstance(split, Split):
        return split
    if split in ["train", "validation", "test"]:
        return Split(name=split, type=split)  # type: ignore[arg-type]
    else:
        return Split(name=split, type="validation")


def _parse_secrets_header(request: Request) -> dict[str, str]:
    raw = request.headers.get("X-Secrets")
    if not raw:
        return {}
    payload = json.loads(base64.b64decode(raw))
    return {k: v["value"] for k, v in payload.items()}


async def extract_sid(request: Request) -> str:
    x_session_id = request.headers.get("X-Session-ID")
    if not x_session_id:
        raise HTTPException(400, "X-Session-ID header is required")
    return x_session_id.strip()


class _LoggingRoute(APIRoute):
    def get_route_handler(self):
        original = super().get_route_handler()
        async def handler(request: Request):
            start = time.monotonic()
            try:
                response = await original(request)
                logger.info(
                    "request_handled",
                    httpRequest={"latency": f"{time.monotonic() - start:.6f}s", "status": response.status_code},
                )
                return response
            except HTTPException as exc:
                logger.warning(
                    "request_rejected",
                    detail=exc.detail,
                    httpRequest={"latency": f"{time.monotonic() - start:.6f}s", "status": exc.status_code},
                )
                raise
            except Exception as exc:
                logger.exception(
                    "request_error",
                    error_type=type(exc).__name__,
                    error=str(exc),
                    httpRequest={"latency": f"{time.monotonic() - start:.6f}s"},
                )
                raise
        return handler


class RequestContextMiddleware:
    """Binds request context (session_id, method, path) to structlog contextvars."""

    def __init__(self, app: ASGIApp):
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] == "http":
            structlog.contextvars.clear_contextvars()
            headers = dict(scope.get("headers", []))
            sid = headers.get(b"x-session-id", b"").decode()
            structlog.contextvars.bind_contextvars(
                session_id=sid,
                method=scope.get("method", ""),
                path=scope.get("path", ""),
            )
        await self.app(scope, receive, send)


class ErrorHandlingMiddleware:
    """Catches unhandled exceptions and returns error details based on configured verbosity."""

    def __init__(self, app: ASGIApp, return_errors: Literal["none", "exception", "stacktrace"] = "stacktrace"):
        self.app = app
        self.return_errors = return_errors

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        try:
            await self.app(scope, receive, send)
        except HTTPException:
            raise
        except Exception as exc:
            logger.exception("unhandled_exception", error_type=type(exc).__name__, error=str(exc))

            if self.return_errors == "none":
                detail = "Internal Server Error"
            elif self.return_errors == "exception":
                detail = f"{type(exc).__name__}: {exc}"
            else:
                detail = traceback.format_exc()

            response = JSONResponse(status_code=500, content={"detail": detail})
            await response(scope, receive, send)


async def _delete_session(
    sid: str,
    active_envs: dict[str, Optional[Environment]],
    last_ping: dict[str, float],
    setup_tasks: dict[str, asyncio.Task],
    ready: dict[str, asyncio.Event],
    setup_errors: dict[str, Exception],
):
    last_ping.pop(sid, None) # stop TTL

    task = setup_tasks.pop(sid, None) # cancel setup if it's running
    if task and not task.done():
        task.cancel()
        with suppress(asyncio.CancelledError):
            await task

    evt = ready.get(sid) # set error for waiters
    if evt and not evt.is_set():
        setup_errors[sid] = HTTPException(410, "Session deleted")
        evt.set()
    ready.pop(sid, None)

    env = active_envs.pop(sid, None) # teardown env
    if env:
        try:
            await maybe_await(env.teardown())
        except Exception as exc:
            logger.exception("Environment teardown failed", session_id=sid, environment_name=env.name(), error_type=type(exc).__name__, error=str(exc))

    setup_errors.pop(sid, None) # clear error for waiters


class RedirectToDefaultEnvMiddleware:
    def __init__(self, app: ASGIApp, env_classes: dict, root_paths: set):
        self.app = app
        self.env_classes = env_classes
        self.root_paths = root_paths

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] == "http":
            path = scope.get("path", "")
            first_segment = path.strip("/").split("/", 1)[0].lower()

            if (
                first_segment not in self.root_paths
                and first_segment not in self.env_classes
                and self.env_classes
            ):
                first_env = next(iter(self.env_classes.keys()))
                query = scope.get("query_string", b"").decode()
                new_path = f"/{first_env}{path}"
                if query:
                    new_path += f"?{query}"
                response = RedirectResponse(url=new_path, status_code=308)
                await response(scope, receive, send)
                return

        await self.app(scope, receive, send)

class Server:
    def __init__(
        self,
        environments: list[type[Environment]],
        return_errors: Literal["none", "exception", "stacktrace"] = "exception",
    ) -> None:
        """Environment hosting server.

        The first environment in the provided list is the default and is served at /.
        ``return_errors`` controls how unhandled exceptions are surfaced in 500 responses:
        "none" returns an opaque message, "exception" includes the exception string,
        and "stacktrace" includes the full traceback.
        """
        if not environments:
            raise ValueError("Server requires at least one environment to be provided.")

        self._env_classes: dict[str, type[Environment]] = _get_env_map(environments)

        # Validate each environment
        for env_name, env_cls in self._env_classes.items():
            # Check that at least one tool is defined
            tools = env_cls.list_tools().tools
            if not tools:
                raise ValueError(f"Environment '{env_name}' has no tools defined. Add at least one @tool method.")

            # Check that at least one split is defined
            splits = env_cls._list_splits_cached()
            if not splits:
                raise ValueError(f"Environment '{env_name}' has no splits defined. Implement list_splits() to return at least one split.")

        self._active_envs: dict[str, Optional[Environment]] = {}
        self._last_ping: dict[str, float] = {}

        self._setup_tasks: dict[str, asyncio.Task] = {}
        self._ready: dict[str, asyncio.Event] = {}
        self._setup_errors: dict[str, Exception] = {}

        self._reaper_task: Optional[asyncio.Task] = None

        async def await_environment_ready(sid: str) -> Environment:
            evt = self._ready.get(sid)
            if evt is None:
                raise HTTPException(404, "Active environment not found")

            if evt.is_set(): # fast path
                err = self._setup_errors.get(sid)
                if err:
                    raise err
                env = self._active_envs.get(sid)
                if env is None:
                    raise HTTPException(410, "Session deleted")
                return env

            await evt.wait()
            err = self._setup_errors.get(sid, None)
            if err:
                raise err
            env = self._active_envs.get(sid)
            if env is None:
                raise HTTPException(410, "Session deleted")
            return env

        async def reaper_coro(reaper_interval: float=5, stale_threshold: float=900):
            while True:
                await asyncio.sleep(reaper_interval)
                now = time.monotonic()
                stale = [sid for sid, ts in list(self._last_ping.items()) if now - ts > stale_threshold]
                for sid in stale:
                    await _delete_session(
                        sid,
                        self._active_envs,
                        self._last_ping,
                        self._setup_tasks,
                        self._ready,
                        self._setup_errors,
                    )

        @asynccontextmanager
        async def lifespan(app: FastAPI):
            self._reaper_task = asyncio.create_task(reaper_coro())
            yield # Runs up to here on startup
            if self._reaper_task:
                self._reaper_task.cancel()
                try:
                    await self._reaper_task
                except asyncio.CancelledError:
                    pass
            await asyncio.gather(*[
                _delete_session(
                    sid,
                    self._active_envs,
                    self._last_ping,
                    self._setup_tasks,
                    self._ready,
                    self._setup_errors,
                )
                for sid in list(self._setup_tasks.keys())
            ])

        app = FastAPI(lifespan=lifespan)
        app.router.route_class = _LoggingRoute

        @app.get("/health")
        async def health():
            return {"status": "ok"}

        @app.get("/version")
        async def version():
            return {"version": __version__}

        @app.post("/ping")
        async def ping(sid: str = Depends(extract_sid)):
            if sid not in self._active_envs:
                raise HTTPException(404, "Active environment not found")
            self._last_ping[sid] = time.monotonic()
            return {"status": "ok"}

        @app.get("/list_environments")
        async def get_envs():
            return list(self._env_classes.keys())

        @app.get("/{env_name}/tools")
        async def list_tools(env_name: str):
            env_cls = _get_env_cls(self._env_classes, env_name)
            return env_cls.list_tools().model_dump()

        @app.get("/{env_name}/splits")
        async def list_splits(env_name: str):
            env_cls = _get_env_cls(self._env_classes, env_name)
            return [_convert_to_split(split).model_dump() for split in env_cls._list_splits_cached()]

        @app.post("/{env_name}/tasks")
        async def list_tasks(env_name: str, list_tasks: ListTasks):
            env_cls = _get_env_cls(self._env_classes, env_name)
            split_names = [split.name if isinstance(split, Split) else split for split in env_cls._list_splits_cached()]
            if list_tasks.split not in split_names:
                raise HTTPException(status_code=400, detail="Invalid split")
            try:
                tasks = await maybe_await(env_cls.list_tasks(list_tasks.split))
            except NotImplementedError as e:
                raise HTTPException(status_code=400, detail="list_tasks is not supported for this environment. Use the index-based API (num_tasks + get_task) instead.") from e
            return {"tasks": tasks, "env_name": env_name}
        
        @app.post("/{env_name}/num_tasks")
        async def num_tasks(env_name: str, num_tasks: NumTasks):
            env_cls = _get_env_cls(self._env_classes, env_name)
            split_names = [split.name if isinstance(split, Split) else split for split in env_cls._list_splits_cached()]
            if num_tasks.split not in split_names:
                raise HTTPException(status_code=400, detail="Invalid split")
            return {"num_tasks": await env_cls.num_tasks(num_tasks.split)}
        
        @app.post("/{env_name}/task")
        async def get_task(env_name: str, get_task: GetTask):
            env_cls = _get_env_cls(self._env_classes, env_name)
            split_names = [split.name if isinstance(split, Split) else split for split in env_cls._list_splits_cached()]
            if get_task.split not in split_names:
                raise HTTPException(status_code=400, detail="Invalid split")
            try:
                task = await env_cls.get_task(get_task.split, get_task.index)
            except IndexError as e:
                raise HTTPException(status_code=400, detail="Invalid index") from e
            return {"task": task}

        @app.post("/{env_name}/task_range")
        async def get_task_range(env_name: str, get_task_range: GetTaskRange):
            env_cls = _get_env_cls(self._env_classes, env_name)
            split_names = [split.name if isinstance(split, Split) else split for split in env_cls._list_splits_cached()]
            if get_task_range.split not in split_names:
                raise HTTPException(status_code=400, detail="Invalid split")
            try:
                tasks = await env_cls.get_task_range(get_task_range.split, get_task_range.start, get_task_range.stop)
            except IndexError as e:
                raise HTTPException(status_code=400, detail="Invalid index range") from e
            return {"tasks": tasks}

        @app.post("/create")
        async def create_environment(request: Request, create_session: CreateSession, sid: str = Depends(extract_sid)):
            if sid in self._active_envs:
                raise HTTPException(status_code=400, detail="Session already exists")

            # Resolve env_name, defaulting to the first registered environment
            env_name = create_session.env_name
            if env_name is None:
                env_name = next(iter(self._env_classes.keys()))
            env_cls = _get_env_cls(self._env_classes, env_name)

            # Resolve task_spec from split/index if needed
            if create_session.task_spec is not None:
                task_spec = create_session.task_spec
            else:
                split_names = [_convert_to_split(s).name for s in env_cls._list_splits_cached()]
                assert create_session.split is not None
                assert create_session.index is not None
                if create_session.split not in split_names:
                    raise HTTPException(status_code=400, detail="Invalid split")
                try:
                    task_spec = await env_cls.get_task(create_session.split, create_session.index)
                except IndexError as e:
                    raise HTTPException(status_code=400, detail="Invalid index") from e

            self._active_envs[sid] = None
            self._last_ping[sid] = time.monotonic()
            self._ready[sid] = asyncio.Event()

            secrets = _parse_secrets_header(request)

            async def perform_setup():
                setup_start = time.monotonic()
                try:
                    env = env_cls(task_spec=task_spec, secrets=secrets)
                    await maybe_await(env.setup())
                    self._active_envs[sid] = env
                    duration_ms = (time.monotonic() - setup_start) * 1000
                    logger.info("setup_completed", session_id=sid, env_name=create_session.env_name, duration_ms=duration_ms)
                except Exception as e:
                    duration_ms = (time.monotonic() - setup_start) * 1000
                    logger.exception("setup_failed", session_id=sid, env_name=create_session.env_name, duration_ms=duration_ms, error_type=type(e).__name__, error=str(e))
                    self._setup_errors[sid] = e
                finally:
                    evt = self._ready.get(sid)
                    if evt:
                        evt.set()

            if sid not in self._setup_tasks:
                self._setup_tasks[sid] = asyncio.create_task(perform_setup())

            logger.info("session_created", env_name=create_session.env_name)
            return {"sid": sid}

        async def require_existing_session(
            sid: str = Depends(extract_sid),
        ):
            self._last_ping[sid] = time.monotonic()
            return await await_environment_ready(sid)


        @app.post("/delete")
        async def delete_environment(sid: str = Depends(extract_sid)):
            if sid in self._setup_tasks:
                await _delete_session(
                    sid,
                    self._active_envs,
                    self._last_ping,
                    self._setup_tasks,
                    self._ready,
                    self._setup_errors,
                )
                logger.info("session_deleted")
            return {"sid": sid}

        @app.post("/create_session")
        async def create_session():
            sid = str(uuid.uuid4())
            async def _stream():
                yield {"event": "task_id", "data": sid}
                yield {"event": "end", "data": ""}
            return EventSourceResponse(
                _stream(),
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                },
            )
        
        @app.post("/delete_session")
        async def delete_session(sid: str = Depends(extract_sid)):
            return {"sid": sid}

        @app.get("/{env_name}/task_tools")
        async def task_tools(env_name: str, env: Environment = Depends(require_existing_session)):
            shared = env.list_tools()
            task_specific = env.list_task_tools()
            combined = ListToolsOutput(tools=shared.tools + task_specific.tools)
            return combined.model_dump()

        @app.post("/{env_name}/call")
        async def call_tool(request: Request, env_name: str, tool_call: ToolCall, env: Environment = Depends(require_existing_session)):
            logger.info("tool_call_started", tool=tool_call.name, env_name=env_name)

            async def call_tool_coro():
                res = await env._call_tool(tool_call.name, tool_call.input)
                return res.model_dump_json(indent=None)

            return EventSourceResponse(
                sse_task_stream(
                    lambda: call_tool_coro(),
                    request,
                    task_id=tool_call.task_id,
                ),
                ping=10,
                media_type="text/event-stream",
                headers={
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    "X-Accel-Buffering": "no",
                }
            )

        @app.get("/{env_name}/prompt")
        async def get_prompt(request: Request, env: Environment = Depends(require_existing_session)) -> Blocks:
            return (await maybe_await(env.get_prompt()))

        # After all routes are defined, before adding the middleware:
        root_paths = set()
        for route in app.routes:
            path = getattr(route, 'path', '')
            segments = path.strip('/').split('/', 1)
            first_segment = segments[0] if segments else ''
            # Root-level paths don't start with a path parameter like {env_name}
            if not first_segment.startswith('{'):
                root_paths.add(first_segment.lower())

        self.app = app
        self.app.add_middleware(
            RedirectToDefaultEnvMiddleware,
            env_classes=self._env_classes,
            root_paths=root_paths
        )
        self.app.add_middleware(RequestContextMiddleware)
        self.app.add_middleware(ErrorHandlingMiddleware, return_errors=return_errors)

    def run(self, host: str = "0.0.0.0", port: int = 8080) -> None:
        """Start the server."""
        setup_logging()
        logger.info("server_starting", version=__version__, host=host, port=port)
        if OPENREWARD_USE_STRUCTURED_LOGS:
            uvicorn.run(self.app, host=host, port=port, timeout_keep_alive=60,
                        log_config=None, access_log=False)
        else:
            uvicorn.run(self.app, host=host, port=port, timeout_keep_alive=60,
                        access_log=False)



from __future__ import annotations

import json
from typing import Literal, Optional
from pydantic import BaseModel, Field

from openreward.environments import Toolset, tool, ToolOutput, TextBlock


# ===== Pydantic Parameter Models =====

class ReadAllParams(BaseModel):
    file_path: str = Field(..., description="Path to PowerPoint file in sandbox (e.g., /orwd_data/presentation.pptx)")


class ReadSlidesParams(BaseModel):
    file_path: str = Field(..., description="Path to PowerPoint file in sandbox")
    slide_indices: Optional[list[int]] = Field(None, description="List of slide indices to read (0-based). If None, reads all slides.")


class CreatePresentationParams(BaseModel):
    file_path: str = Field(..., description="Output path for new presentation in sandbox")
    title: Optional[str] = Field(None, description="Optional title text for first slide")


class DeletePresentationParams(BaseModel):
    file_path: str = Field(..., description="Path to presentation file to delete in sandbox")


class AddSlideParams(BaseModel):
    file_path: str = Field(..., description="Path to presentation in sandbox")
    layout: Literal["title", "title_content", "section_header", "two_content", "comparison", "title_only", "blank", "content_with_caption", "picture_with_caption"] = Field(..., description="Slide layout type")
    title: Optional[str] = Field(None, description="Title text for the slide")
    content: Optional[str] = Field(None, description="Body/content text for the slide")


class EditSlideParams(BaseModel):
    file_path: str = Field(..., description="Path to presentation in sandbox")
    slide_index: int = Field(..., description="Zero-based slide index to edit")
    title: Optional[str] = Field(None, description="New title text (updates title shape/placeholder)")
    content: Optional[str] = Field(None, description="New body content (updates content placeholder)")


class AddImageParams(BaseModel):
    file_path: str = Field(..., description="Path to presentation in sandbox")
    slide_index: int = Field(..., description="Zero-based slide index to add image to")
    image_path: str = Field(..., description="Path to image file in sandbox")
    left: float = Field(..., description="Left position in inches")
    top: float = Field(..., description="Top position in inches")
    width: Optional[float] = Field(None, description="Width in inches (None = auto)")
    height: Optional[float] = Field(None, description="Height in inches (None = auto)")


class ModifyImageParams(BaseModel):
    file_path: str = Field(..., description="Path to presentation in sandbox")
    slide_index: int = Field(..., description="Zero-based slide index")
    image_index: int = Field(..., description="Zero-based index of image on slide")
    new_image_path: Optional[str] = Field(None, description="Path to replacement image (None = keep existing)")
    width: Optional[float] = Field(None, description="New width in inches")
    height: Optional[float] = Field(None, description="New height in inches")


class ChartSeriesData(BaseModel):
    name: str = Field(..., description="Series name")
    values: list[float] = Field(..., description="Data values for series")


class AddChartParams(BaseModel):
    file_path: str = Field(..., description="Path to presentation in sandbox")
    slide_index: int = Field(..., description="Zero-based slide index")
    chart_type: Literal["column", "bar", "line", "pie"] = Field(..., description="Type of chart")
    categories: list[str] = Field(..., description="Category labels (x-axis)")
    series: list[ChartSeriesData] = Field(..., description="Data series for chart")
    left: float = Field(..., description="Left position in inches")
    top: float = Field(..., description="Top position in inches")
    width: float = Field(..., description="Width in inches")
    height: float = Field(..., description="Height in inches")


class AddTableParams(BaseModel):
    file_path: str = Field(..., description="Path to presentation in sandbox")
    slide_index: int = Field(..., description="Zero-based slide index")
    rows: int = Field(..., description="Number of rows")
    cols: int = Field(..., description="Number of columns")
    data: list[list[str]] = Field(..., description="2D array of cell contents [row][col]")
    left: float = Field(..., description="Left position in inches")
    top: float = Field(..., description="Top position in inches")
    width: float = Field(..., description="Width in inches")
    height: float = Field(..., description="Height in inches")


class ListImagesParams(BaseModel):
    file_path: str = Field(..., description="Path to presentation in sandbox")
    slide_index: Optional[int] = Field(None, description="Specific slide index (None = all slides)")


# ===== PowerPoint Toolset Class =====

class PowerPointToolset(Toolset):
    """Toolset providing PowerPoint manipulation tools via python-pptx executed in sandbox"""

    @tool
    async def powerpoint_read_all(self, params: ReadAllParams) -> ToolOutput:
        """Read complete presentation structure including metadata and all slide content"""
        script = f'''
import json
from pptx import Presentation

try:
    prs = Presentation("{params.file_path}")

    slides_data = []
    for idx, slide in enumerate(prs.slides):
        text_parts = []
        for shape in slide.shapes:
            if hasattr(shape, "has_text_frame") and shape.has_text_frame:
                for paragraph in shape.text_frame.paragraphs:
                    for run in paragraph.runs:
                        if run.text.strip():
                            text_parts.append(run.text)

        slides_data.append({{
            "index": idx,
            "layout": slide.slide_layout.name,
            "text": "\\n".join(text_parts),
            "shape_count": len(slide.shapes),
        }})

    result = {{
        "success": True,
        "slide_count": len(prs.slides),
        "width": prs.slide_width / 914400,
        "height": prs.slide_height / 914400,
        "slides": slides_data,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error reading presentation: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            # Build summary
            summary = f"Presentation: {params.file_path}\n"
            summary += f"Slides: {result['slide_count']}\n"
            summary += f'Dimensions: {result["width"]:.2f}" x {result["height"]:.2f}"\n\n'
            summary += "Slide Overview:\n"
            for s in result['slides'][:5]:
                preview = s['text'][:60].replace("\n", " ")
                summary += f"  [{s['index']}] {s['layout']}: {preview}...\n"
            if result['slide_count'] > 5:
                summary += f"  ... and {result['slide_count'] - 5} more slides\n"

            return ToolOutput(
                blocks=[TextBlock(text=summary)],
                metadata={
                    "file_path": params.file_path,
                    "slide_count": result['slide_count'],
                    "dimensions": {"width": result['width'], "height": result['height']},
                    "slides": result['slides'],
                },
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_read_slides(self, params: ReadSlidesParams) -> ToolOutput:
        """Read text content from specific slides or all slides"""
        indices_str = json.dumps(params.slide_indices) if params.slide_indices else "None"

        script = f'''
import json
from pptx import Presentation

try:
    prs = Presentation("{params.file_path}")
    indices = {indices_str}

    if indices is None:
        indices = list(range(len(prs.slides)))

    slides_data = []
    for idx in indices:
        if idx < 0 or idx >= len(prs.slides):
            slides_data.append({{"index": idx, "error": "Slide index out of range"}})
            continue

        slide = prs.slides[idx]
        text_parts = []
        for shape in slide.shapes:
            if hasattr(shape, "has_text_frame") and shape.has_text_frame:
                for paragraph in shape.text_frame.paragraphs:
                    for run in paragraph.runs:
                        if run.text.strip():
                            text_parts.append(run.text)

        slides_data.append({{
            "index": idx,
            "layout": slide.slide_layout.name,
            "text": "\\n".join(text_parts),
        }})

    result = {{"success": True, "slides": slides_data}}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            summary = f"Read {len(result['slides'])} slide(s) from {params.file_path}\n\n"
            for s in result['slides']:
                if "error" in s:
                    summary += f"Slide {s['index']}: {s['error']}\n"
                else:
                    text_preview = s['text'].replace("\n", " ")
                    summary += f"Slide {s['index']} ({s['layout']}):\n{text_preview}"

            return ToolOutput(
                blocks=[TextBlock(text=summary)],
                metadata={"file_path": params.file_path, "slides": result['slides']},
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_create(self, params: CreatePresentationParams) -> ToolOutput:
        """Create a new PowerPoint presentation with optional title slide"""
        title_escaped = params.title.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n') if params.title else ""

        script = f'''
import json
from pptx import Presentation

try:
    prs = Presentation()

    if "{title_escaped}":
        title_slide_layout = prs.slide_layouts[0]
        slide = prs.slides.add_slide(title_slide_layout)
        title = slide.shapes.title
        title.text = "{title_escaped}"

    prs.save("{params.file_path}")

    result = {{"success": True, "slide_count": len(prs.slides)}}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Presentation created successfully at {params.file_path}")],
                metadata={
                    "file_path": params.file_path,
                    "slide_count": result['slide_count'],
                },
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_delete(self, params: DeletePresentationParams) -> ToolOutput:
        """Delete a PowerPoint presentation file from sandbox"""
        output, exit_code = await self.sandbox.run(f"rm {params.file_path}", max_bytes=1_000_000)

        if exit_code == 0:
            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Presentation deleted successfully: {params.file_path}")],
                metadata={"file_path": params.file_path, "deleted": True},
                reward=0.0,
                finished=False,
            )
        else:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Failed to delete: {output}")],
                metadata={"file_path": params.file_path, "deleted": False, "error": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_add_slide(self, params: AddSlideParams) -> ToolOutput:
        """Add a new slide to presentation with specified layout and content"""
        layout_map = {
            "title": 0,
            "title_content": 1,
            "section_header": 2,
            "two_content": 3,
            "comparison": 4,
            "title_only": 5,
            "blank": 6,
            "content_with_caption": 7,
            "picture_with_caption": 8,
        }
        layout_idx = layout_map.get(params.layout, 1)

        # Escape quotes and newlines for Python string literal
        title_escaped = params.title.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n') if params.title else ""
        content_escaped = params.content.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n') if params.content else ""

        script = f'''
import json
from pptx import Presentation

try:
    prs = Presentation("{params.file_path}")
    slide_layout = prs.slide_layouts[{layout_idx}]
    slide = prs.slides.add_slide(slide_layout)

    if "{title_escaped}" and slide.shapes.title:
        slide.shapes.title.text = "{title_escaped}"

    if "{content_escaped}":
        for shape in slide.placeholders:
            if shape.placeholder_format.idx == 1:
                shape.text = "{content_escaped}"
                break

    prs.save("{params.file_path}")

    result = {{
        "success": True,
        "slide_index": len(prs.slides) - 1,
        "slide_count": len(prs.slides),
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Slide added successfully at index {result['slide_index']}")],
                metadata={
                    "file_path": params.file_path,
                    "slide_index": result['slide_index'],
                    "slide_count": result['slide_count'],
                    "layout": params.layout,
                },
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_edit_slide(self, params: EditSlideParams) -> ToolOutput:
        """Edit text content of an existing slide"""
        title_escaped = params.title.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n') if params.title else None
        content_escaped = params.content.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n') if params.content else None

        script = f'''
import json
from pptx import Presentation

try:
    prs = Presentation("{params.file_path}")

    if {params.slide_index} < 0 or {params.slide_index} >= len(prs.slides):
        raise ValueError(f"Slide index {params.slide_index} out of range")

    slide = prs.slides[{params.slide_index}]

    if "{title_escaped}" != "None" and slide.shapes.title:
        slide.shapes.title.text = "{title_escaped}"

    if "{content_escaped}" != "None":
        for shape in slide.placeholders:
            if shape.placeholder_format.idx == 1:
                shape.text = "{content_escaped}"
                break

    prs.save("{params.file_path}")

    result = {{"success": True}}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Slide {params.slide_index} updated successfully")],
                metadata={"file_path": params.file_path, "slide_index": params.slide_index},
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_add_table(self, params: AddTableParams) -> ToolOutput:
        """Add a table to a slide with specified data"""
        data_json = json.dumps(params.data)

        script = f'''
import json
from pptx import Presentation
from pptx.util import Inches

try:
    prs = Presentation("{params.file_path}")
    slide = prs.slides[{params.slide_index}]

    data = {data_json}
    table_shape = slide.shapes.add_table(
        {params.rows}, {params.cols},
        Inches({params.left}), Inches({params.top}),
        Inches({params.width}), Inches({params.height})
    )
    table = table_shape.table

    for row_idx, row_data in enumerate(data):
        for col_idx, cell_value in enumerate(row_data):
            table.cell(row_idx, col_idx).text = str(cell_value)

    prs.save("{params.file_path}")

    result = {{"success": True}}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Table ({params.rows}x{params.cols}) added to slide {params.slide_index}")],
                metadata={
                    "file_path": params.file_path,
                    "slide_index": params.slide_index,
                    "rows": params.rows,
                    "cols": params.cols,
                },
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_add_image(self, params: AddImageParams) -> ToolOutput:
        """Add an image to a slide"""
        width_param = f"width=Inches({params.width})" if params.width else ""
        height_param = f"height=Inches({params.height})" if params.height else ""
        size_params = f", {width_param}" if width_param else ""
        size_params += f", {height_param}" if height_param else ""

        script = f'''
import json
from pptx import Presentation
from pptx.util import Inches

try:
    prs = Presentation("{params.file_path}")
    slide = prs.slides[{params.slide_index}]

    pic = slide.shapes.add_picture(
        "{params.image_path}",
        Inches({params.left}),
        Inches({params.top}){size_params}
    )

    width_inches = pic.width / 914400
    height_inches = pic.height / 914400

    prs.save("{params.file_path}")

    result = {{
        "success": True,
        "width": width_inches,
        "height": height_inches,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error"), "image_added": False},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Image added to slide {params.slide_index}")],
                metadata={
                    "file_path": params.file_path,
                    "slide_index": params.slide_index,
                    "image_added": True,
                    "dimensions": {"width": result['width'], "height": result['height']},
                },
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output, "image_added": False},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_add_chart(self, params: AddChartParams) -> ToolOutput:
        """Add a chart to a slide"""
        categories_json = json.dumps(params.categories)
        series_json = json.dumps([{"name": s.name, "values": s.values} for s in params.series])

        chart_type_map = {
            "column": "XL_CHART_TYPE.COLUMN_CLUSTERED",
            "bar": "XL_CHART_TYPE.BAR_CLUSTERED",
            "line": "XL_CHART_TYPE.LINE",
            "pie": "XL_CHART_TYPE.PIE",
        }
        chart_type_code = chart_type_map.get(params.chart_type, "XL_CHART_TYPE.COLUMN_CLUSTERED")

        script = f'''
import json
from pptx import Presentation
from pptx.util import Inches
from pptx.enum.chart import XL_CHART_TYPE
from pptx.chart.data import ChartData

try:
    prs = Presentation("{params.file_path}")
    slide = prs.slides[{params.slide_index}]

    chart_data = ChartData()
    chart_data.categories = {categories_json}

    series_list = {series_json}
    for series in series_list:
        chart_data.add_series(series["name"], series["values"])

    chart_shape = slide.shapes.add_chart(
        {chart_type_code},
        Inches({params.left}), Inches({params.top}),
        Inches({params.width}), Inches({params.height}),
        chart_data
    )

    prs.save("{params.file_path}")

    result = {{
        "success": True,
        "series_count": len(series_list),
        "category_count": len({categories_json}),
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ {params.chart_type.capitalize()} chart added to slide {params.slide_index}")],
                metadata={
                    "file_path": params.file_path,
                    "slide_index": params.slide_index,
                    "chart_type": params.chart_type,
                    "series_count": result['series_count'],
                    "category_count": result['category_count'],
                },
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_list_images(self, params: ListImagesParams) -> ToolOutput:
        """List all images in presentation or specific slide"""
        slide_filter = f"[{params.slide_index}]" if params.slide_index is not None else "None"

        script = f'''
import json
from pptx import Presentation
from pptx.enum.shapes import MSO_SHAPE_TYPE

try:
    prs = Presentation("{params.file_path}")
    images = []

    slide_indices = {slide_filter} if {slide_filter} != None else list(range(len(prs.slides)))

    for slide_idx in slide_indices:
        slide = prs.slides[slide_idx]
        img_idx = 0
        for shape in slide.shapes:
            if shape.shape_type == MSO_SHAPE_TYPE.PICTURE:
                images.append({{
                    "slide_index": slide_idx,
                    "image_index": img_idx,
                    "width": shape.width / 914400,
                    "height": shape.height / 914400,
                    "left": shape.left / 914400,
                    "top": shape.top / 914400,
                }})
                img_idx += 1

    result = {{"success": True, "images": images}}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            images = result['images']
            summary = f"Found {len(images)} image(s) in {params.file_path}\n\n"
            for img in images:
                summary += f"Slide {img['slide_index']}, Image {img['image_index']}: "
                summary += f'{img["width"]:.2f}"x{img["height"]:.2f}" at ({img["left"]:.2f}", {img["top"]:.2f}")\n'

            return ToolOutput(
                blocks=[TextBlock(text=summary)],
                metadata={"file_path": params.file_path, "images": images},
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def powerpoint_modify_image(self, params: ModifyImageParams) -> ToolOutput:
        """Resize or replace an existing image on a slide"""
        width_code = f"image_shape.width = int(Inches({params.width}))" if params.width else ""
        height_code = f"image_shape.height = int(Inches({params.height}))" if params.height else ""

        script = f'''
import json
from pptx import Presentation
from pptx.util import Inches
from pptx.enum.shapes import MSO_SHAPE_TYPE

try:
    prs = Presentation("{params.file_path}")
    slide = prs.slides[{params.slide_index}]

    image_shapes = [s for s in slide.shapes if s.shape_type == MSO_SHAPE_TYPE.PICTURE]

    if {params.image_index} < 0 or {params.image_index} >= len(image_shapes):
        raise ValueError(f"Image index {params.image_index} out of range")

    image_shape = image_shapes[{params.image_index}]

    {width_code}
    {height_code}

    new_width = image_shape.width / 914400
    new_height = image_shape.height / 914400

    prs.save("{params.file_path}")

    result = {{
        "success": True,
        "width": new_width,
        "height": new_height,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Image modified on slide {params.slide_index}")],
                metadata={
                    "file_path": params.file_path,
                    "slide_index": params.slide_index,
                    "image_index": params.image_index,
                    "new_dimensions": {"width": result['width'], "height": result['height']},
                },
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )



import asyncio
from typing import Any, Dict, Optional

from aiohttp import (ClientError, ClientResponseError, ClientSession,
                     ContentTypeError)
from tenacity import (AsyncRetrying, retry_if_exception, stop_after_attempt,
                      wait_exponential)


def should_retry_on_500(exception: BaseException) -> bool:
    """Retry only on 5xx server errors"""
    if isinstance(exception, ClientResponseError):
        return 500 <= exception.status < 600
    return False


async def make_request(
    client: ClientSession,
    url: str,
    method: str,
    data: Dict[str, Any],
    headers: Dict[str, str],
    body: Optional[Dict[str, Any]],
    max_retries: int,
    backoff_base: float,
    backoff_factor: float,
    backoff_cap: float,
) -> Dict[str, Any]:

    retryer = AsyncRetrying(
        retry=retry_if_exception(should_retry_on_500),
        wait=wait_exponential(multiplier=backoff_base, exp_base=backoff_factor, max=backoff_cap),
        stop=stop_after_attempt(1 + max_retries),
        reraise=True,
    )

    async for attempt in retryer:
        with attempt:
            async with client.request(
                method.upper(),
                url,
                params=data,
                json=body,
                headers=headers,
            ) as resp:
                text = await resp.text()
                try:
                    resp.raise_for_status()
                except ClientResponseError as e:
                    raise ClientResponseError(
                        request_info=e.request_info,
                        history=e.history,
                        status=e.status,
                        message=text,
                        headers=e.headers,
                    )
                try:
                    return await resp.json()
                except (ContentTypeError, ValueError):
                    text = await resp.text()
                    return {"status": resp.status, "text": text}

    raise RuntimeError("Retry exhausted without return or exception")



import asyncio
import json
import os
import uuid
from typing import Any, Dict, Optional

import atexit
import click
import structlog
import sys
import time
from anthropic.types import MessageParam as AnthropicMessageParam
from google.genai import types as gdm_types

from openreward.models import (
    LogMessageEvent,
    RolloutConfig,
    RolloutStartedEvent,
    SendLoopConfig,
    ShutdownEvent,
)
from openreward.http_client import make_request
from .background import start_background_worker
from .serializers.ant import serialize_anthropic_message
from .serializers.base import UploadType, base_to_normalized
from .serializers.gdm import serialize_gdm_message
from .serializers.models import NormalizedEvent
from .serializers.oai_completions import (
    OpenAIChatMessage,
    openai_completions_to_normalized,
)
from .serializers.oai_responses import serialize_openai_response

from openreward.log_utils import get_logger as _get_logger
logger = _get_logger("openreward.rollouts.rollout")

_json_logger = structlog.wrap_logger(
    structlog.PrintLogger(sys.stdout),
    processors=[
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.add_log_level,
        structlog.processors.JSONRenderer(),
    ],
)

def _rollout_logging_format() -> str:
    # Read at call time so os.environ mutations before first use are respected.
    # Valid values: "pretty", "structured". Defaults to "pretty" if unset.
    return os.getenv("OPENREWARD_ROLLOUT_LOGGING_FORMAT", "pretty").lower()


async def _fetch_environment_id(send_loop_config: SendLoopConfig, owner: str, env_name: str) -> Optional[str]:
    from aiohttp import ClientSession
    from openreward._version import USER_AGENT
    async with ClientSession(base_url=send_loop_config.base_url) as session:
        result = await make_request(
            client=session,
            url=f"/v1/environments/{owner}/{env_name}",
            method="GET",
            data={},
            headers={"User-Agent": USER_AGENT},
            body=None,
            max_retries=send_loop_config.max_retries,
            backoff_base=send_loop_config.backoff_base,
            backoff_factor=send_loop_config.backoff_factor,
            backoff_cap=send_loop_config.backoff_cap,
        )
        return result.get("id")

class RolloutAPI:
    def __init__(self,
        send_loop_config: SendLoopConfig,
        shutdown_timeout: float = 10.0,
        web_base_url: str = "https://openreward.ai",
    ):
        self.send_loop_config = send_loop_config
        self.shutdown_timeout = shutdown_timeout
        self.web_base_url = web_base_url

        self._loop, self._in, self._thread = start_background_worker(send_loop_config)
        self._closed = False

        atexit.register(self._atexit_handler)

    def create(
        self,
        run_name: str,
        rollout_name: Optional[str] = None,
        environment: Optional[str] = None,
        split: Optional[str] = None,
        metadata: Optional[dict] = None,
        task_spec: Optional[Dict[str, Any]] = None,
        print_messages: bool = False,
    ) -> "Rollout":
        config = RolloutConfig(
            run_name=run_name,
            rollout_name=rollout_name,
            environment=environment,
            split=split,
            metadata=metadata,
            task_spec=task_spec,
        )
        return Rollout(config=config, _in=self._in, _loop=self._loop, web_base_url=self.web_base_url, send_loop_config=self.send_loop_config, print_messages=print_messages)

    def close(self):
        if self._closed:
            return
        self._closed = True
        if self._thread.is_alive():
            self._loop.call_soon_threadsafe(self._in.put_nowait, ShutdownEvent())
            self._thread.join(timeout=self.shutdown_timeout)
            if self._thread.is_alive():
                print("openreward: shutdown timed out, some rollout data may not have been uploaded")

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def _atexit_handler(self):
        # Best effort if user forgot
        try:
            self.close()
        except Exception:
            pass

class Rollout:
    def __init__(self, config: RolloutConfig, _in: asyncio.Queue, _loop: asyncio.AbstractEventLoop, web_base_url: str = "https://openreward.ai", send_loop_config: Optional[SendLoopConfig] = None, print_messages: bool = False):
        self.config = config
        self.event_id = str(uuid.uuid4())
        self._in = _in
        self._loop = _loop
        self.logged_messages = 0
        self.print_messages = print_messages

        self.environment_id: Optional[str] = None
        if send_loop_config and config.environment and "/" in config.environment:
            owner, env_name = config.environment.split("/", 1)
            try:
                future = asyncio.run_coroutine_threadsafe(
                    _fetch_environment_id(send_loop_config, owner, env_name),
                    self._loop,
                )
                self.environment_id = future.result(timeout=10)
            except Exception:
                logger.debug("failed to fetch environment_id", environment=config.environment)

        self._loop.call_soon_threadsafe(self._in.put_nowait, RolloutStartedEvent.from_config(
            event_id=self.event_id,
            timestamp=int(time.time() * 1000),
            config=config,
            environment_id=self.environment_id,
        ))

        url = f"{web_base_url.rstrip('/')}/rollout/{self.event_id}"
        if print_messages:
            if _rollout_logging_format() == "pretty":
                prefix = click.style("openreward:", bold=True, fg="blue")
                recording = click.style("Recording rollout", fg=(247, 230, 204))
                styled_url = click.style(url, fg="blue", underline=True)
                click.echo(f"{prefix} \U0001f680 {recording} {styled_url}", err=True)
            elif _rollout_logging_format() == "structured":
                _json_logger.info("rollout_started", rollout_id=self.event_id, url=url)

    def _print_message(self, normalized_event: NormalizedEvent, reward: Optional[float] = None, is_finished: Optional[bool] = False):
        BLUE = "\033[34m"
        GREEN = "\033[32m"
        YELLOW = "\033[33m"
        CYAN = "\033[36m"
        DIM = "\033[2m"
        BOLD = "\033[1m"
        RESET = "\033[0m"

        def _print_line(char="─"):
            print(f"{DIM}{char * 80}{RESET}")

        def _decode_newlines(s: str) -> str:
            return s.replace("\\n", "\n")

        t = normalized_event.type

        if t == "reasoning":
            _print_line()
            print(f"{BOLD}{DIM}💭 Reasoning{RESET}")
            _print_line()
            if normalized_event.summary:
                print(f"\n  {DIM}{_decode_newlines(normalized_event.summary)}{RESET}")
            print()
        elif t == "tool_call":
            _print_line()
            print(f"{YELLOW}⚡ Tool Call:{RESET} {BOLD}{normalized_event.name}{RESET}  {DIM}(call_id: {normalized_event.call_id}){RESET}")
            _print_line()
            if normalized_event.content:
                try:
                    args_str = json.dumps(json.loads(normalized_event.content), indent=2)
                except (json.JSONDecodeError, TypeError):
                    args_str = _decode_newlines(normalized_event.content)
                for line in args_str.splitlines():
                    print(line)
            print()
        elif t == "tool_result":
            _print_line()
            status = f"{GREEN}✓ done{RESET}" if is_finished else f"{CYAN}… continuing{RESET}"
            reward_str = f"  {DIM}reward={reward}{RESET}" if reward is not None else ""
            print(f"{GREEN}↩ Tool Result:{RESET}  {DIM}(call_id: {normalized_event.call_id}){RESET}  [{status}{reward_str}]")
            _print_line()
            if normalized_event.content:
                try:
                    args_str = json.dumps(json.loads(normalized_event.content), indent=2)
                except (json.JSONDecodeError, TypeError):
                    args_str = _decode_newlines(normalized_event.content)
                for line in args_str.splitlines():
                    print(line)
            print()
        elif t == "assistant_message":
            _print_line()
            print(f"{BOLD}{BLUE}🤖 Assistant{RESET}")
            _print_line()
            if normalized_event.content:
                print(f"\n{_decode_newlines(normalized_event.content)}")
            print()
        elif t == "user_message":
            _print_line()
            print(f"{BOLD}👤 User{RESET}")
            _print_line()
            if normalized_event.content:
                print(f"\n{_decode_newlines(normalized_event.content)}")
            print()
        elif t == "system_message":
            _print_line()
            print(f"{DIM}⚙ System{RESET}")
            _print_line()
            if normalized_event.content:
                print(f"\n{DIM}{_decode_newlines(normalized_event.content)}{RESET}")
            print()

    def _log_message(self, normalized_event: NormalizedEvent, reward: Optional[float] = None, is_finished: Optional[bool] = False, metadata: Optional[dict] = None):

        if self.print_messages:
            if _rollout_logging_format() == "pretty":
                self._print_message(normalized_event, reward, is_finished)
            elif _rollout_logging_format() == "structured":
                _json_logger.info(
                    "rollout_message",
                    rollout_id=self.event_id,
                    type=normalized_event.type,
                    name=normalized_event.name,
                    call_id=normalized_event.call_id,
                    summary=normalized_event.summary,
                    reward=reward,
                    is_finished=is_finished,
                )

        if is_finished is None:
            is_finished = False

        self._loop.call_soon_threadsafe(self._in.put_nowait, LogMessageEvent(
            eventId = str(uuid.uuid4()),
            timestamp=int(time.time() * 1000),
            index=self.logged_messages,
            rolloutEventId=self.event_id,
            environment_id=self.environment_id,
            type=normalized_event.type,
            content=normalized_event.content,
            contentReference=normalized_event.content_reference,
            summary=normalized_event.summary,
            name=normalized_event.name,
            callId=normalized_event.call_id,
            reward=reward,
            is_finished=is_finished,
            metadata=metadata or {},
        ))
        self.logged_messages += 1

    def log(
        self,
        message: UploadType,
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        normalized_event = base_to_normalized([message])[0]
        self._log_message(normalized_event, reward, is_finished, metadata)

    def log_openai_completions(
        self,
        message: OpenAIChatMessage,
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        normalized_event = openai_completions_to_normalized([message])[0]
        self._log_message(normalized_event, reward, is_finished, metadata)

    def log_openai_response(
        self,
        message,  # Accept both Response and ResponseInputItemParam
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        # Handle Response objects by iterating over output
        messages_to_log = []
        if hasattr(message, 'output'):
            messages_to_log = list(message.output)
        else:
            messages_to_log = [message]

        for msg in messages_to_log:
            event = serialize_openai_response(msg)
            if event:  # Skip None (shouldn't happen with fixes, but safe)
                self._log_message(event, reward, is_finished, metadata)

    def log_anthropic_message(
        self,
        message: AnthropicMessageParam,
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        for event in serialize_anthropic_message(message):
            self._log_message(event, reward, is_finished, metadata)

    def log_gdm_message(
        self,
        message: gdm_types.Content,
        reward: Optional[float] = None,
        is_finished: Optional[bool] = False,
        metadata: Optional[dict] = None,
    ):
        for event in serialize_gdm_message(message):
            self._log_message(event, reward, is_finished, metadata)



"""Command line utilities for managing OpenReward environments."""
from __future__ import annotations

import argparse
import re
from dataclasses import dataclass, field
from importlib import resources
from pathlib import Path
from string import Template
from typing import Callable, Mapping


@dataclass(frozen=True)
class TemplateFiles:
    """Collection of scaffold files for an environment."""

    dockerfile: str
    server_py: str
    extra_files: Mapping[str, str] = field(default_factory=dict)


def _pascal_case(name: str) -> str:
    parts = re.split(r"[^0-9a-zA-Z]+", name)
    cleaned = "".join(part.capitalize() for part in parts if part)
    if not cleaned:
        cleaned = "Environment"
    if cleaned[0].isdigit():
        cleaned = f"Env{cleaned}"
    return cleaned


def _load_template_file(template_name: str, filename: str) -> str:
    package = f"openreward.templates.{template_name}"
    file = resources.files(package).joinpath(filename)
    return file.read_text(encoding="utf-8")


_README_TEMPLATE = """\
# {env_name}

[![OpenReward Environment](https://img.shields.io/badge/%E2%AD%90%20OpenReward-Environment-f7e6cc)](https://openreward.ai/YourOrg/{env_name})

## Description

<!-- 2-4 sentences: what does this environment evaluate? What domain does it cover?
     Mention the verification method and the source dataset. -->

## Capabilities

<!-- Bulleted list of skills or capabilities the environment tests. -->

- Capability 1
- Capability 2
- Capability 3

## Compute Requirements

<!-- Does the environment need a sandbox? GPU? Extra memory?
     If it runs without a sandbox, say so. -->

## License

<!-- License that covers the use of the environment, as well as any underlying data or code it depends on.
     Link to the license text. If the environment and its dependencies have different licenses, list both. -->

## Tasks

<!-- How many tasks are there? What splits (train/test)?
     Briefly describe the structure of each task. -->

## Reward Structure

<!-- How are rewards computed?
     Sparse or dense? Binary or continuous?
     Is verification programmatic or does it use an LLM grader? Does it use rubrics?
     If there are multiple validation steps, list them in order. -->

## Data

<!-- Where does the task data come from? How is it stored?
     Link to the source dataset if applicable. -->

## Tools

<!-- Explain the tools available to the agent. -->

## Time Horizon

<!-- Single-turn or multi-turn?
     If multi-turn, roughly how many tool calls does a typical task require? -->

## Environment Difficulty

<!-- Solve rates, baseline model performance, or other difficulty statistics. -->

## Other Environment Requirements

<!-- Any external dependencies: API keys, secrets, third-party services. -->

## Safety

<!-- Does the agent have access to external systems (network, file system, APIs)?
     Are there dual-use risks in the domain (e.g. chemistry, cybersecurity)?
     Is there a possibility of goal misspecification?
     What mitigations are in place? -->

## Citations

<!-- BibTeX entries for the environment itself and any underlying datasets or papers. -->

```bibtex
@dataset{{YourCitation,
  author    = {{Your Name or Team}},
  title     = {{{env_name}}},
  year      = {{2026}},
  publisher = {{OpenReward}},
  url       = {{https://openreward.ai/YourOrg/{env_name}}}
}}
```
"""


def _render_readme(env_name: str) -> str:
    return _README_TEMPLATE.format(env_name=env_name)


def _basic_template(env_name: str) -> TemplateFiles:
    class_name = _pascal_case(env_name)
    dockerfile = _load_template_file("basic", "Dockerfile")
    server_template = Template(_load_template_file("basic", "server.py.tmpl"))
    server_py = server_template.substitute(CLASS_NAME=class_name)
    extra_files = {
        "requirements.txt": _load_template_file("basic", "requirements.txt"),
        "README.md": _render_readme(env_name),
    }
    return TemplateFiles(dockerfile=dockerfile, server_py=server_py, extra_files=extra_files)


def _sandbox_template(env_name: str) -> TemplateFiles:
    """Create a sandbox template with Docker-in-Docker support."""
    dockerfile = _load_template_file("sandbox", "Dockerfile")
    server_py = _load_template_file("sandbox", "server.py")
    extra_files = {
        "requirements.txt": _load_template_file("sandbox", "requirements.txt"),
        "sandbox_env.py": _load_template_file("sandbox", "sandbox_env.py"),
        "README.md": _render_readme(env_name),
    }
    return TemplateFiles(dockerfile=dockerfile, server_py=server_py, extra_files=extra_files)


TemplateFactory = Callable[[str], TemplateFiles]


TEMPLATES: Mapping[str, TemplateFactory] = {
    "basic": _basic_template,
    "sandbox": _sandbox_template,
}


def _render_template(template_name: str, env_name: str) -> TemplateFiles:
    generator = TEMPLATES.get(template_name)
    if generator is None:
        raise ValueError(f"Unknown template '{template_name}'. Available templates: {', '.join(TEMPLATES)}")
    return generator(env_name)


def _write_file(path: Path, contents: str) -> None:
    if path.exists():
        raise FileExistsError(f"{path} already exists")
    path.write_text(contents, encoding="utf-8")


def command_init(environment: str, template: str) -> None:
    target_dir = Path.cwd() / environment
    target_dir.mkdir(parents=True, exist_ok=False)

    files = _render_template(template, environment)

    _write_file(target_dir / "Dockerfile", files.dockerfile)
    _write_file(target_dir / "server.py", files.server_py)
    for relative, contents in files.extra_files.items():
        _write_file(target_dir / relative, contents)


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="orwd", description="OpenReward CLI utilities")
    subparsers = parser.add_subparsers(dest="command", required=True)

    init_parser = subparsers.add_parser(
        "init",
        help="Initialize a new environment template",
    )
    init_parser.add_argument("environment", help="Name of the environment to scaffold")
    init_parser.add_argument(
        "--template",
        default="basic",
        choices=sorted(TEMPLATES.keys()),
        help="Scaffold template to use (default: %(default)s)",
    )
    init_parser.set_defaults(func=lambda args: command_init(args.environment, args.template))

    return parser


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    try:
        args.func(args)
    except FileExistsError as exc:
        parser.error(str(exc))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Word document manipulation toolset using python-docx."""

from __future__ import annotations

import json
from typing import Literal, Optional, Any
from pydantic import BaseModel, Field

from openreward.environments import tool, ToolOutput, TextBlock
from openreward.environments.toolset import Toolset


# ===== Pydantic Parameter Models =====

class CreateDocumentParams(BaseModel):
    file_path: str = Field(..., description="Path where the new document will be created")
    title: str | None = Field(None, description="Document title metadata")
    author: str | None = Field(None, description="Document author metadata")


class GetOverviewParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")


class ReadDocumentParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")
    include_tables: bool = Field(True, description="Include table content in output")
    max_paragraphs: int | None = Field(None, description="Limit number of paragraphs returned")


class ReadImageParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")
    image_index: int | None = Field(None, description="Specific image index to read (if None, returns all images)")
    include_data: bool = Field(True, description="Include base64-encoded image data in response")


class AddContentParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")
    text: str = Field(..., description="Text content to add")
    content_type: Literal["paragraph", "heading"] = Field("paragraph", description="Type of content")
    heading_level: int | None = Field(None, description="Heading level (1-9) if content_type is heading")
    style: str | None = Field(None, description="Paragraph style name")
    position: Literal["end", "start"] = Field("end", description="Where to insert content")


class AddImageParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")
    image_path: str = Field(..., description="Path to the image file to insert")
    width_inches: float | None = Field(None, description="Image width in inches")
    height_inches: float | None = Field(None, description="Image height in inches")
    position: Literal["end", "start"] = Field("end", description="Where to insert image")


class EditContentParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")
    paragraph_index: int = Field(..., description="Zero-based paragraph index")
    new_text: str = Field(..., description="New text content")
    append: bool = Field(False, description="Append to existing text instead of replacing")


class DeleteContentParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")
    paragraph_index: int = Field(..., description="Zero-based paragraph index to delete")


class ModifyImageParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")
    image_index: int = Field(..., description="Zero-based image index")
    width_inches: float | None = Field(None, description="New width in inches")
    height_inches: float | None = Field(None, description="New height in inches")


class ApplyFormattingParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")
    paragraph_index: int = Field(..., description="Zero-based paragraph index")
    bold: bool | None = Field(None, description="Apply bold formatting")
    italic: bool | None = Field(None, description="Apply italic formatting")
    underline: bool | None = Field(None, description="Apply underline formatting")
    font_name: str | None = Field(None, description="Font family name (e.g., 'Arial')")
    font_size: int | None = Field(None, description="Font size in points")
    color: str | None = Field(None, description="Hex color code (e.g., 'FF0000' for red)")


class DeleteDocumentParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document to delete")


class SearchDocumentParams(BaseModel):
    file_path: str = Field(..., description="Path to the Word document")
    search_text: str = Field(..., description="Text to search for")
    case_sensitive: bool = Field(False, description="Case-sensitive search")
    search_in_tables: bool = Field(True, description="Include table content in search")
    max_results: int | None = Field(None, description="Limit number of results")


# ===== Word Toolset Class =====

class WordToolset(Toolset):
    """Toolset providing Word document manipulation tools via python-docx executed in sandbox"""

    @tool
    async def word_create_document(self, params: CreateDocumentParams) -> ToolOutput:
        """Create a new Word document with optional metadata"""
        title_escaped = params.title.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n') if params.title else ""
        author_escaped = params.author.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n') if params.author else ""

        script = f'''
import json
from docx import Document

try:
    doc = Document()

    # Set metadata if provided
    if "{title_escaped}":
        doc.core_properties.title = "{title_escaped}"
    if "{author_escaped}":
        doc.core_properties.author = "{author_escaped}"

    doc.save("{params.file_path}")

    result = {{
        "success": True,
        "file_path": "{params.file_path}",
        "title": "{title_escaped}" if "{title_escaped}" else None,
        "author": "{author_escaped}" if "{author_escaped}" else None,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Document created successfully at {params.file_path}")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_delete_document(self, params: DeleteDocumentParams) -> ToolOutput:
        """Delete a Word document file from sandbox"""
        output, exit_code = await self.sandbox.run(f"rm {params.file_path}", max_bytes=1_000_000)

        if exit_code == 0:
            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Document deleted successfully: {params.file_path}")],
                metadata={"file_path": params.file_path, "deleted": True},
                reward=0.0,
                finished=False,
            )
        else:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Failed to delete: {output}")],
                metadata={"file_path": params.file_path, "deleted": False, "error": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_get_document_overview(self, params: GetOverviewParams) -> ToolOutput:
        """Retrieve document structure and metadata including paragraph count, table count, and image count"""
        script = f'''
import json
from docx import Document

try:
    doc = Document("{params.file_path}")

    # Count paragraphs
    paragraph_count = len(doc.paragraphs)

    # Count tables
    table_count = len(doc.tables)

    # Count images (inline shapes)
    image_count = 0
    for rel in doc.part.rels.values():
        if "image" in rel.target_ref:
            image_count += 1

    # Get metadata
    props = doc.core_properties
    metadata = {{
        "title": props.title if props.title else "",
        "author": props.author if props.author else "",
        "created": str(props.created) if props.created else "",
        "modified": str(props.modified) if props.modified else "",
    }}

    result = {{
        "success": True,
        "file_path": "{params.file_path}",
        "paragraph_count": paragraph_count,
        "table_count": table_count,
        "image_count": image_count,
        "metadata": metadata,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            # Build summary
            summary = f"Document: {params.file_path}\n"
            summary += f"Paragraphs: {result['paragraph_count']}\n"
            summary += f"Tables: {result['table_count']}\n"
            summary += f"Images: {result['image_count']}\n\n"
            summary += "Metadata:\n"
            for key, value in result['metadata'].items():
                if value:
                    summary += f"  {key.capitalize()}: {value}\n"

            return ToolOutput(
                blocks=[TextBlock(text=summary)],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_read_document_content(self, params: ReadDocumentParams) -> ToolOutput:
        """Extract all text content with paragraph structure from the document"""
        max_para = params.max_paragraphs if params.max_paragraphs else 999999

        script = f'''
import json
from docx import Document

try:
    doc = Document("{params.file_path}")

    paragraphs = []
    for idx, para in enumerate(doc.paragraphs):
        if idx >= {max_para}:
            break

        paragraphs.append({{
            "index": idx,
            "text": para.text,
            "style": para.style.name if para.style else "Normal",
        }})

    # Optionally include table content
    tables = []
    if {params.include_tables}:
        for table_idx, table in enumerate(doc.tables):
            table_data = []
            for row in table.rows:
                row_data = [cell.text for cell in row.cells]
                table_data.append(row_data)
            tables.append({{
                "index": table_idx,
                "rows": len(table.rows),
                "cols": len(table.columns),
                "data": table_data,
            }})

    result = {{
        "success": True,
        "file_path": "{params.file_path}",
        "paragraph_count": len(paragraphs),
        "paragraphs": paragraphs,
        "table_count": len(tables),
        "tables": tables,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            # Build summary
            summary = f"Document content from {params.file_path}\n"
            summary += f"Paragraphs: {result['paragraph_count']}\n"
            if result['table_count'] > 0:
                summary += f"Tables: {result['table_count']}\n"
            summary += "\n"

            # Show first few paragraphs
            for para in result['paragraphs']:
                text_preview = para['text'] if para['text'] else "(empty)"
                summary += f"{text_preview}\n\n"

            return ToolOutput(
                blocks=[TextBlock(text=summary)],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_read_image(self, params: ReadImageParams) -> ToolOutput:
        """Extract image data and metadata from document"""
        script = f'''
import json
import base64
from docx import Document

try:
    doc = Document("{params.file_path}")

    images = []

    # Iterate through all parts to find images
    for rel in doc.part.rels.values():
        if "image" in rel.target_ref:
            image_part = rel.target_part
            image_data = image_part.blob  # Binary data

            # Extract format from content_type (e.g., 'image/png' -> 'png')
            image_format = image_part.content_type.split('/')[-1]

            image_info = {{
                "index": len(images),
                "format": image_format,
                "size_bytes": len(image_data),
            }}

            # Include base64 data if requested
            if {params.include_data}:
                image_base64 = base64.b64encode(image_data).decode('utf-8')
                image_info["data"] = image_base64

            images.append(image_info)

    # Filter by image_index if specified
    if {params.image_index} is not None:
        if {params.image_index} < 0 or {params.image_index} >= len(images):
            print(json.dumps({{
                "success": False,
                "error": f"Image index {params.image_index} out of range (document has {{len(images)}} images)"
            }}))
        else:
            images = [images[{params.image_index}]]

    result = {{
        "success": True,
        "file_path": "{params.file_path}",
        "image_count": len(images),
        "images": images,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            # Build summary
            summary = f"Found {result['image_count']} image(s) in {params.file_path}\n\n"
            for img in result['images']:
                summary += f"Image {img['index']}: {img['format'].upper()} ({img['size_bytes']} bytes)\n"
                if params.include_data and 'data' in img:
                    summary += f"  Base64 data: {len(img['data'])} characters\n"

            return ToolOutput(
                blocks=[TextBlock(text=summary)],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_add_content_text(self, params: AddContentParams) -> ToolOutput:
        """Add paragraphs or headings to document"""
        text_escaped = params.text.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
        style_escaped = params.style.replace('\\', '\\\\').replace('"', '\\"') if params.style else ""

        # Build the content addition logic
        if params.content_type == "heading":
            if not params.heading_level or params.heading_level < 1 or params.heading_level > 9:
                return ToolOutput(
                    blocks=[TextBlock(text="❌ Error: heading_level must be between 1 and 9 for heading content")],
                    metadata={"error": "Invalid heading_level"},
                    reward=0.0,
                    finished=False,
                )
            add_code = f'new_para = doc.add_heading("{text_escaped}", level={params.heading_level})'
        else:
            add_code = f'new_para = doc.add_paragraph("{text_escaped}")'
            if params.style:
                add_code += f'\n    new_para.style = "{style_escaped}"'

        # Handle positioning (start vs end)
        if params.position == "start":
            insert_code = f'''
    # Insert at start by inserting before first paragraph
    if len(doc.paragraphs) > 0:
        p = doc.paragraphs[0]._element
        p.getparent().insert(0, new_para._element)
    else:
        {add_code}
'''
        else:
            insert_code = f'    {add_code}'

        script = f'''
import json
from docx import Document

try:
    doc = Document("{params.file_path}")

{insert_code}

    doc.save("{params.file_path}")

    result = {{
        "success": True,
        "file_path": "{params.file_path}",
        "content_type": "{params.content_type}",
        "position": "{params.position}",
        "paragraph_count": len(doc.paragraphs),
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            content_label = f"Heading {params.heading_level}" if params.content_type == "heading" else "Paragraph"
            return ToolOutput(
                blocks=[TextBlock(text=f"✅ {content_label} added at {params.position} of document")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_add_image(self, params: AddImageParams) -> ToolOutput:
        """Insert images into document"""
        width_param = f", width=Inches({params.width_inches})" if params.width_inches else ""
        height_param = f", height=Inches({params.height_inches})" if params.height_inches else ""

        # Handle positioning
        if params.position == "start":
            insert_code = f'''
    # Add image in a new paragraph at the start
    paragraph = doc.paragraphs[0] if len(doc.paragraphs) > 0 else doc.add_paragraph()
    run = paragraph.insert_paragraph_before().add_run()
    picture = run.add_picture("{params.image_path}"{width_param}{height_param})
'''
        else:
            insert_code = f'''
    # Add image at end
    paragraph = doc.add_paragraph()
    run = paragraph.add_run()
    picture = run.add_picture("{params.image_path}"{width_param}{height_param})
'''

        script = f'''
import json
from docx import Document
from docx.shared import Inches

try:
    doc = Document("{params.file_path}")

{insert_code}

    # Get final dimensions
    width_inches = picture.width / 914400
    height_inches = picture.height / 914400

    doc.save("{params.file_path}")

    result = {{
        "success": True,
        "file_path": "{params.file_path}",
        "image_path": "{params.image_path}",
        "position": "{params.position}",
        "width": width_inches,
        "height": height_inches,
    }}
    print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Image added at {params.position} of document ({result['width']:.2f}\" x {result['height']:.2f}\")")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_edit_content_text(self, params: EditContentParams) -> ToolOutput:
        """Modify existing paragraph text"""
        text_escaped = params.new_text.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')

        if params.append:
            update_code = f'para.add_run("\\n{text_escaped}")'
        else:
            update_code = f'para.text = "{text_escaped}"'

        script = f'''
import json
from docx import Document

try:
    doc = Document("{params.file_path}")

    if {params.paragraph_index} < 0 or {params.paragraph_index} >= len(doc.paragraphs):
        print(json.dumps({{
            "success": False,
            "error": f"Paragraph index {params.paragraph_index} out of range (document has {{len(doc.paragraphs)}} paragraphs)"
        }}))
    else:
        para = doc.paragraphs[{params.paragraph_index}]
        old_text = para.text

        {update_code}

        doc.save("{params.file_path}")

        result = {{
            "success": True,
            "file_path": "{params.file_path}",
            "paragraph_index": {params.paragraph_index},
            "old_text": old_text[:100],
            "new_text": para.text[:100],
            "append": {params.append},
        }}
        print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            action = "appended to" if params.append else "updated"
            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Paragraph {params.paragraph_index} {action} successfully\nPreview: {result['new_text']}...")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_delete_content_text(self, params: DeleteContentParams) -> ToolOutput:
        """Remove paragraphs by index"""
        script = f'''
import json
from docx import Document

try:
    doc = Document("{params.file_path}")

    if {params.paragraph_index} < 0 or {params.paragraph_index} >= len(doc.paragraphs):
        print(json.dumps({{
            "success": False,
            "error": f"Paragraph index {params.paragraph_index} out of range (document has {{len(doc.paragraphs)}} paragraphs)"
        }}))
    else:
        para = doc.paragraphs[{params.paragraph_index}]
        deleted_text = para.text

        # Delete paragraph element
        p = para._element
        p.getparent().remove(p)

        doc.save("{params.file_path}")

        result = {{
            "success": True,
            "file_path": "{params.file_path}",
            "paragraph_index": {params.paragraph_index},
            "deleted_text": deleted_text[:100],
            "remaining_paragraphs": len(doc.paragraphs),
        }}
        print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Paragraph {params.paragraph_index} deleted successfully\nDeleted text: {result['deleted_text']}...")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_modify_image(self, params: ModifyImageParams) -> ToolOutput:
        """Update image properties (size)"""
        width_code = f"picture.width = Inches({params.width_inches})" if params.width_inches else ""
        height_code = f"picture.height = Inches({params.height_inches})" if params.height_inches else ""

        script = f'''
import json
from docx import Document
from docx.shared import Inches

try:
    doc = Document("{params.file_path}")

    # Try using inline_shapes if available
    if hasattr(doc, 'inline_shapes'):
        if {params.image_index} < 0 or {params.image_index} >= len(doc.inline_shapes):
            print(json.dumps({{
                "success": False,
                "error": f"Image index {params.image_index} out of range (document has {{len(doc.inline_shapes)}} images)"
            }}))
        else:
            picture = doc.inline_shapes[{params.image_index}]

            {width_code}
            {height_code}

            new_width = picture.width / 914400
            new_height = picture.height / 914400

            doc.save("{params.file_path}")

            result = {{
                "success": True,
                "file_path": "{params.file_path}",
                "image_index": {params.image_index},
                "new_width": new_width,
                "new_height": new_height,
            }}
            print(json.dumps(result))
    else:
        print(json.dumps({{
            "success": False,
            "error": "Image modification requires python-docx with inline_shapes support"
        }}))

except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Image {params.image_index} modified successfully ({result['new_width']:.2f}\" x {result['new_height']:.2f}\")")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_apply_formatting(self, params: ApplyFormattingParams) -> ToolOutput:
        """Apply text formatting to paragraph runs"""
        # Build formatting code
        format_lines = []
        if params.bold is not None:
            format_lines.append(f"run.font.bold = {str(params.bold)}")
        if params.italic is not None:
            format_lines.append(f"run.font.italic = {str(params.italic)}")
        if params.underline is not None:
            format_lines.append(f"run.font.underline = {str(params.underline)}")
        if params.font_name:
            font_escaped = params.font_name.replace('\\', '\\\\').replace('"', '\\"')
            format_lines.append(f'run.font.name = "{font_escaped}"')
        if params.font_size:
            format_lines.append(f"run.font.size = Pt({params.font_size})")
        if params.color:
            # Convert hex color (e.g., "FF0000") to RGB
            format_lines.append(f"run.font.color.rgb = RGBColor(int('{params.color}'[0:2], 16), int('{params.color}'[2:4], 16), int('{params.color}'[4:6], 16))")

        format_code = "\n                ".join(format_lines) if format_lines else "pass"

        script = f'''
import json
from docx import Document
from docx.shared import Pt, RGBColor

try:
    doc = Document("{params.file_path}")

    if {params.paragraph_index} < 0 or {params.paragraph_index} >= len(doc.paragraphs):
        print(json.dumps({{
            "success": False,
            "error": f"Paragraph index {params.paragraph_index} out of range (document has {{len(doc.paragraphs)}} paragraphs)"
        }}))
    else:
        para = doc.paragraphs[{params.paragraph_index}]

        # Apply formatting to all runs in the paragraph
        for run in para.runs:
            if run.text.strip():
                {format_code}

        doc.save("{params.file_path}")

        result = {{
            "success": True,
            "file_path": "{params.file_path}",
            "paragraph_index": {params.paragraph_index},
            "formatting_applied": {{
                "bold": {params.bold},
                "italic": {params.italic},
                "underline": {params.underline},
                "font_name": "{params.font_name}" if "{params.font_name}" else None,
                "font_size": {params.font_size},
                "color": "{params.color}" if "{params.color}" else None,
            }},
        }}
        print(json.dumps(result))
except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            return ToolOutput(
                blocks=[TextBlock(text=f"✅ Formatting applied to paragraph {params.paragraph_index} successfully")],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )

    @tool
    async def word_search_document(self, params: SearchDocumentParams) -> ToolOutput:
        """Search for text within document paragraphs and tables"""
        search_text_escaped = params.search_text.replace('\\', '\\\\').replace('"', '\\"')
        max_results_repr = repr(params.max_results)

        script = f'''
import json
from docx import Document
import re

try:
    doc = Document("{params.file_path}")

    # Prepare search pattern
    flags = 0 if {params.case_sensitive} else re.IGNORECASE
    pattern = re.compile(re.escape("{search_text_escaped}"), flags=flags)

    results = []
    max_results = {max_results_repr}

    # Search in paragraphs
    for para_idx, para in enumerate(doc.paragraphs):
        text = para.text
        if not text:
            continue

        for match in pattern.finditer(text):
            start = max(0, match.start() - 50)
            end = min(len(text), match.end() + 50)
            context = text[start:end]

            results.append({{
                "location_type": "paragraph",
                "paragraph_index": para_idx,
                "start_pos": match.start(),
                "end_pos": match.end(),
                "context": context,
                "style": para.style.name if para.style else "Normal"
            }})

            if max_results and len(results) >= max_results:
                break

        if max_results and len(results) >= max_results:
            break

    # Search in tables if enabled and not at max_results
    if {params.search_in_tables} and (not max_results or len(results) < max_results):
        for table_idx, table in enumerate(doc.tables):
            for row_idx, row in enumerate(table.rows):
                for col_idx, cell in enumerate(row.cells):
                    text = cell.text
                    if not text:
                        continue

                    for match in pattern.finditer(text):
                        start = max(0, match.start() - 50)
                        end = min(len(text), match.end() + 50)
                        context = text[start:end]

                        results.append({{
                            "location_type": "table",
                            "table_index": table_idx,
                            "row_index": row_idx,
                            "col_index": col_idx,
                            "start_pos": match.start(),
                            "end_pos": match.end(),
                            "context": context
                        }})

                        if max_results and len(results) >= max_results:
                            break

                    if max_results and len(results) >= max_results:
                        break

                if max_results and len(results) >= max_results:
                    break

            if max_results and len(results) >= max_results:
                break

    result = {{
        "success": True,
        "search_text": "{search_text_escaped}",
        "total_results": len(results),
        "results": results
    }}
    print(json.dumps(result))

except Exception as e:
    print(json.dumps({{"success": False, "error": str(e)}}))
'''

        cmd = f"python3 << 'PYSCRIPT_EOF'\n{script}\nPYSCRIPT_EOF"
        output, exit_code = await self.sandbox.run(cmd, max_bytes=1_000_000)
        try:
            result = json.loads(output)

            if not result.get("success"):
                return ToolOutput(
                    blocks=[TextBlock(text=f"❌ Error: {result.get('error')}")],
                    metadata={"error": result.get("error")},
                    reward=0.0,
                    finished=False,
                )

            count = result["total_results"]
            display_text = f"✅ Found {count} occurrence(s) of '{params.search_text}'"

            if count > 0:
                display_text += "\n\nMatches:"
                for i, match in enumerate(result["results"][:5], 1):
                    if match["location_type"] == "paragraph":
                        display_text += f"\n{i}. Paragraph {match['paragraph_index']} ({match['style']}): ...{match['context']}..."
                    else:
                        display_text += f"\n{i}. Table {match['table_index']} [Row {match['row_index']}, Col {match['col_index']}]: ...{match['context']}..."

                if count > 5:
                    display_text += f"\n\n... and {count - 5} more"

            return ToolOutput(
                blocks=[TextBlock(text=display_text)],
                metadata=result,
                reward=0.0,
                finished=False,
            )
        except json.JSONDecodeError:
            return ToolOutput(
                blocks=[TextBlock(text=f"❌ Error parsing output: {output}")],
                metadata={"error": "JSON decode failed", "output": output},
                reward=0.0,
                finished=False,
            )



from .http import (
    AuthenticationError,
    HeartbeatTimeoutError,
    MaxRetriesError,
    _finalize_session,
    _raise_for_status,
    _raise_for_status_with_auth,
    request_retryable,
    resumable_sse,
)
from .ping import ErrorResponse
from .session import BaseAsyncSession, SessionTerminatedError

__all__ = [
    "AuthenticationError",
    "BaseAsyncSession",
    "ErrorResponse",
    "HeartbeatTimeoutError",
    "MaxRetriesError",
    "SessionTerminatedError",
    "_finalize_session",
    "_raise_for_status",
    "_raise_for_status_with_auth",
    "request_retryable",
    "resumable_sse",
]
